#! python
'''
UCL positronium spectroscopy blog, https://antimattergravity.com.

Usage:
    >>> from positronium import antimattergravity

Try also:
    >>> import antigravity
'''
import webbrowser

webbrowser.open("https://antimattergravity.com/")



#! python
'''
The Bohr model of positronium
'''
from __future__ import print_function, division
import numpy as np
from . import constants

def energy(n1=1, n2=float('inf'), **kwargs):
    '''
    Calculate the interval between energy levels n1 and n2
    according to the Bohr model, i.e., the Rydberg formula.

    kwargs:
        unit:
            J, eV, meV, ueV, au (Hartree),          [energy]
            Hz, kHz, MHz, GHz, THz,                 [frequency]
            m, cm, mm, um, nm, A, pm, fm,           [vacuum wavelength]
            m^-1, cm^-1.                            [wavenumber]

    defaults:
        n1 = 1
        n2 = infinity
        unit='eV'

    '''
    unit = kwargs.get('unit', 'eV')
    interval = 0.25 * (np.subtract(np.power(n1, -2.0),
                                   np.power(n2, -2.0)))
    if (unit == 'au') or (unit == 'Hartree'):
        return interval
    elif unit in constants.au_energy:
        return constants.au_energy[unit](interval)
    else:
        raise KeyError(unit + ' is not recognised as a suitable unit. See' + \
                              ' docstring for unit list.')

def radius(n=1, **kwargs):
    '''
    Return the Bohr radius for positronium (2 * a_0).

    kwargs:
        unit:
            m, cm, mm, um, nm, A, pm, fm, (SI)
            au (Bohr).

    defaults:
        n = 1
        unit = 'm'

    '''
    unit = kwargs.get('unit', 'm')
    rad = (2.0 * np.power(n, 2.0))
    if (unit == 'au') or (unit == 'Bohr'):
        return rad
    elif unit in constants.au_distance:
        return constants.au_distance[unit](rad)
    else:
        raise KeyError(unit + ' is not recognised as a suitable unit. See' + \
                              ' docstring for unit list.')



#! python
'''
Atomic state of positronium
'''
from __future__ import print_function, division
from math import factorial
import string
from scipy.special import sph_harm, hyp1f1
import numpy as np
from . import constants, Ferrell

TAU_0 = 3.0*constants.hbar / (2.0 * constants.alpha**5.0 * \
        constants.reduced_mass_Ps * constants.c**2.0)

class Ps(object):
    """ Ps atomic state described by the quantum numbers n, l, m, S and J.
    """
    def __init__(self, n, l=0, m=0, S=0, J=None):
        self.n = n
        self.l = l
        self.m = m
        self.S = S
        if J is None:
            self.J = self.l + self.S
        else:
            self.J = J
        self.check_quantum_numbers()

    def check_quantum_numbers(self):
        """" Validate chosen quantum numbers. """
        if not isinstance(self.n, (int)) or self.n <= 0:
            raise ValueError("The principal quantum number 'n' must be a " + \
                             "positive integer")
        if not isinstance(self.l, (int)) or self.l < 0 or self.l >= self.n:
            raise ValueError("The orbital angular momentum quantum number 'l' " + \
                             "must be a positive or zero integer, and must be " + \
                             "less than the principal quantum number 'n'.")
        if not isinstance(self.S, (int)) or not (self.S == 0 or self.S == 1):
            raise ValueError("The total spin quantum number 'S' " + \
                             "must be an integer equal to 1 or 0.")
        if not isinstance(self.J, (int)) or not (self.l - self.S) <= self.J <= (self.l + self.S):
            raise ValueError("The total angular momentum quantum number 'J' " + \
                             "must be in the range l - 1 < J < l + 1")
        if self.l == 0 and self.J != self.S:
            raise ValueError("The total angular momentum quantum number 'J' " + \
                             "must be equal to S if l = 0.")
    # ------------------------
    # Schrodinger wavefunction
    # ------------------------

    def Y_lm(self, theta, phi):
        """ The angular part of the wavefunction, Y_lm(theta, phi).

            quantum numbers:
                l, m
        """
        return sph_harm(self.m, self.l, phi, theta)

    def R_nl(self, r):
        """ The radial part of the wavefunction, R_nl(r).
                Quantum Mechanics of one and two electron atoms,
                H. A. Bethe and E. E. Salpeter 1957

                r in units of the Bohr radius, a_0.

            quantum numbers:
                n, l
        """
        rho = 1.0 * r / self.n           # re-scaled rho by 1/2 from hydrogen
        epsilon = 1.0 / self.n
        c1 = np.sqrt(factorial(self.n + self.l) / (2.0 * self.n * \
                     factorial(self.n - self.l - 1))) * \
            1.0/ (factorial(2.0 * self.l + 1.0))
        c2 = np.power(2.0, -3.0/2.0)    # re-normalise again for Ps
        return c1 * c2 * (2* epsilon)**(3.0/2.0) * np.exp(-0.5* rho) * rho**self.l *\
               hyp1f1(-(self.n - self.l - 1), 2 * self.l + 2, rho)

    def wavefunction_nlm(self, r, theta, phi):
        """ wavefunction_nlm(r, theta, phi) = R_nl (r) * Y_lm (theta, phi)

                r in units of the Bohr radius, a_0.

            quantum numbers:
                n, l, m
        """
        return self.R_nl(r) * self.Y_lm(theta, phi)

    # ------
    # energy
    # ------

    def energy(self, unit='J'):
        """ energy levels of positronium according to calculations by
            Richard A. Ferrell (1951) Phys. Rev. 84, 858
            http://dx.doi.org/10.1103/PhysRev.84.858

            This does not include the Lamb shift/ radiative corrections.

            quantum numbers:
                n, l, S, J
        """
        return Ferrell.energy_level(self.n, self.l, self.S, self.J, unit=unit)

    # ----------
    # ionisation
    # ----------

    def ionization_efield(self, unit='V m^-1'):
        """ Ionization in an electric field, for a given n.

            For each value of n the ionization field for the outermost Stark state with
            a negative Stark shift is approximately equal to the classical ionization
            field.

            T. F. Gallagher, Rydberg Atoms (Cambridge University Press, Cambridge,
            England, 1994).

            quantum numbers:
                n

            kwargs:
                unit:
                    V m^-1, V cm^-1      [electric field]

            defaults:
                n = 1
                unit='V cm^-1'
        """
        rescale = {'V m^-1': (lambda x: x),
                   'V cm^-1': (lambda x: 0.01 * x)}
        if unit not in rescale:
            raise KeyError('"' + unit + '" is not recognised as a suitable unit. See' +
                           ' docstring for unit list.')
        else:
            efield = np.power(self.n, -4.0) * 2.0 * constants.Ryd_Ps * \
                               constants.h * constants.c / (constants.e * constants.a_Ps * 9.0)
            return rescale[unit](efield)

    # ---------------
    # radiative decay
    # ---------------

    def radiative_decay(self, unit='Hz'):
        """ A universal formula for the radiative mean rate/ lifetime of hydrogenlike
            states. The formula is accurate to at least 6% for the lowest
            states and to a much higher degree of accuracy for highly excited states.

            Semiclassical estimation of the radiative mean lifetimes of hydrogenlike states
            Hermann Marxer and Larry Spruch
            Phys. Rev. A 43, 1268
            https://dx.doi.org/10.1103/PhysRevA.43.1268

            quantum numbers:
                n, l>0

            kwargs:
                unit:
                    s, ms, us, ns, ps,                      [lifetime]
                    Hz, kHz, MHz, GHz, THz, PHz, EHz,       [rate]

            defaults:
                unit='Hz'
        """
        if self.l == 0:
            raise ValueError('This semiclassical estimate does not work for l=0 states.')
        else:
            rescale = {'s': (lambda x: x),
                       'ms': (lambda x: x*1e3),
                       'us': (lambda x: x*1e6),
                       'ns': (lambda x: x*1e9),
                       'ps': (lambda x: x*1e12),
                       'Hz': (lambda x: 1.0/ x),
                       'kHz': (lambda x: 1.0/ x * 1e-3),
                       'MHz': (lambda x: 1.0/ x * 1e-6),
                       'GHz': (lambda x: 1.0/ x * 1e-9),
                       'THz': (lambda x: 1.0/ x * 1e-12),
                       'PHz': (lambda x: 1.0/ x * 1e-15),
                       'EHz': (lambda x: 1.0/ x * 1e-18)}
            if unit not in rescale:
                raise KeyError('"' + unit + '" is not recognised as a suitable unit. See' +
                               ' docstring for unit list.')
            else:
                lifetime = TAU_0 * np.multiply(np.power(self.n, 3.0),
                                               np.multiply(self.l, np.add(self.l, 1)))
                return rescale[unit](lifetime)

    # ----------------
    # self annihlation
    # ----------------

    def annihilation_rate(self, unit='Hz'):
        """  Calculate the annihilation rate/ lifetime for S states (l=0).

            quantum numbers:
                n, l=0
        """
        if self.l != 0:
            raise ValueError('This only works for l=0 states.')
        else:
            rescale = {'s': (lambda x: x),
                       'ms': (lambda x: x*1e3),
                       'us': (lambda x: x*1e6),
                       'ns': (lambda x: x*1e9),
                       'ps': (lambda x: x*1e12),
                       'Hz': (lambda x: 1.0/ x),
                       'kHz': (lambda x: 1.0/ x * 1e-3),
                       'MHz': (lambda x: 1.0/ x * 1e-6),
                       'GHz': (lambda x: 1.0/ x * 1e-9),
                       'THz': (lambda x: 1.0/ x * 1e-12),
                       'PHz': (lambda x: 1.0/ x * 1e-15),
                       'EHz': (lambda x: 1.0/ x * 1e-18)}
            if unit not in rescale:
                raise KeyError('"' + unit + '" is not recognised as a suitable unit.')
            else:
                if self.S == 0:
                    return constants.tau_pPs * self.n**3.0
                elif self.S == 1:
                    return constants.tau_oPs * self.n**3.0
                else:
                    raise ValueError("The total spin quantum number 'S' must be 0 or 1.")

    # -----
    # LaTeX
    # -----

    def tex(self):
        """ Convert the quantum numbers into a tex string of the form n^{2S + 1}L_{J}

            quantum numbers:
                n, l, S, J

            In IPython use Latex to render, e.g.

            >>> from IPython.display import Latex
            >>> from positronium import Ps
            >>> x = Ps(n=1, l=0, S=1, J=1)
            >>> Latex(x.tex())
        """
        L = ('SPD' + string.ascii_uppercase[5:].replace("J", ""))[self.l]
        return r'$%d^{%d}'%(self.n, 2*self.S + 1) + L + r'_{%d}$'%(self.J)



#! python
'''
Physical constants
'''
from __future__ import print_function, division
import webbrowser
from scipy.constants import m_e, e, c, h, hbar, alpha, Rydberg

class MeasuredValue(float):
    ''' subclass of built-in type float'''
    def __new__(cls, value, uncertainty, unit):
        measured = float.__new__(cls, value)
        measured.value = value
        measured.uncertainty = uncertainty
        measured.unit = unit
        measured.source = None
        measured.url = None
        return measured

    def article(self):
        ''' open url to constant reference'''
        if self.url is not None:
            webbrowser.open(self.url)
        else:
            raise LookupError('The source url has not been specified.')

# Planck constant uncertainty CODATA [J s]
h_uncertainty = 8.1e-42
# mass [kg]
m_Ps = 2.0 * m_e
# reduced mass [kg]
reduced_mass_Ps = m_e/ 2.0
# Rydberg
Ryd_Ps = Rydberg / 2.0
Rydberg_Ps = Rydberg / 2.0
# Bohr radius [m]
a_0 = hbar / (m_e * c * alpha)
# Ps Bohr raidus [m]
a_Ps = 2.0 * a_0

# ground-state decay rate / lifetime
## para-positronium
decay_pPs = MeasuredValue(7990.9e6, 1.7e6, 'Hz')
tau_pPs = MeasuredValue(1.0 / decay_pPs, decay_pPs.uncertainty/ decay_pPs**2.0, 's')
for con in [decay_pPs, tau_pPs]:
    setattr(con, 'source', 'Al-Ramadhan, A. H. and Gidley, D. W. (1994) ' + \
                           'Phys. Rev. Lett. 72, 1632 ')
    setattr(con, 'url', 'http://dx.doi.org/10.1103/PhysRevLett.72.1632')

## ortho-positronium
decay_oPs = MeasuredValue(7.0404e6, 0.0018e6, 'Hz')
tau_oPs = MeasuredValue(1.0 / decay_oPs, decay_oPs.uncertainty/ decay_oPs**2.0, 's')
for con in [decay_oPs, tau_oPs]:
    setattr(con, 'source', 'R. S. Vallery, P. W. Zitzewitz, and D. W. Gidley (2003) ' + \
                           'Phys. Rev. Lett. 90, 203402')
    setattr(con, 'url', 'http://dx.doi.org/10.1103/PhysRevLett.90.203402')

# ground-state hyperfine splitting
nu_hfs = MeasuredValue(203.38910e9, 000.00074e9, 'Hz')
energy_hfs = MeasuredValue(h * nu_hfs,
                           ((h * nu_hfs.uncertainty)**2.0 + (h_uncertainty * nu_hfs)**2.0)**0.5,
                           'J')
for con in [nu_hfs, energy_hfs]:
    setattr(con, 'source', 'M. W. Ritter, P. O. Egan, V. W. Hughes, and ' +
                           'K. A. Woodle (1984) Phys. Rev. A 30, 1331')
    setattr(con, 'url', 'http://dx.doi.org/10.1103/PhysRevA.30.1331')

# 1^3S-2^3S interval
nu_1s2s = MeasuredValue(1233607216.4e6, 3.2e6, 'Hz')
energy_1s2s = MeasuredValue(h * nu_1s2s,
                            ((h * nu_1s2s.uncertainty)**2.0 + (h_uncertainty * nu_1s2s)**2.0)**0.5,
                            'J')
for con in [nu_1s2s, energy_1s2s]:
    setattr(con, 'source', 'Fee, M.S. et al. (1993) Phys. Rev. Lett. 70, 1397')
    setattr(con, 'url', 'http://dx.doi.org/10.1103/PhysRevLett.70.1397')

# rescale atomic units
## energy / energy interval / wavelength/ wavenumbers
au_energy = dict({'J': (lambda x: x * 2.0 * Rydberg * h * c),
                  'eV': (lambda x: x * 2.0 * Rydberg * h * c / e),
                  'meV': (lambda x: 1e3 * x * 2.0 * Rydberg * h * c / e),
                  'ueV': (lambda x: 1e6 * x * 2.0 * Rydberg * h * c / e),
                  'Hz': (lambda x: x * 2.0 * Rydberg * c),
                  'kHz': (lambda x: 1e-3 * x * 2.0 * Rydberg * c),
                  'MHz': (lambda x: 1e-6 * x * 2.0 * Rydberg * c),
                  'GHz': (lambda x: 1e-9 * x * 2.0 * Rydberg * c),
                  'THz': (lambda x: 1e-12 * x * 2.0 * Rydberg * c),
                  'm': (lambda x: 1.0 / (x * 2.0 * Rydberg)),
                  'cm': (lambda x: 1e2 / (x * 2.0 * Rydberg)),
                  'mm': (lambda x: 1e3 / (x * 2.0 * Rydberg)),
                  'um': (lambda x: 1e6 / (x * 2.0 * Rydberg)),
                  'nm': (lambda x: 1e9 / (x * 2.0 * Rydberg)),
                  'A': (lambda x: 1e10 / (x * 2.0 * Rydberg)),
                  'pm': (lambda x: 1e12 / (x * 2.0 * Rydberg)),
                  'fm': (lambda x: 1e15 / (x * 2.0 * Rydberg)),
                  'm^-1': (lambda x: x * 2.0 * Rydberg),
                  'cm^-1': (lambda x: 1e-2 * x * 2.0 * Rydberg)})
## distance
au_distance = dict({'m': (lambda x: x * a_0),
                    'cm': (lambda x: x * a_0 * 1e2),
                    'mm': (lambda x: x * a_0 * 1e3),
                    'um': (lambda x: x * a_0 * 1e6),
                    'nm': (lambda x: x * a_0 * 1e9),
                    'A': (lambda x: x * a_0 * 1e10),
                    'pm': (lambda x: x * a_0 * 1e12),
                    'fm': (lambda x: x * a_0 * 1e15)})



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='positronium',
      version='0.1.6',
      description='python tools pertaining to positronium',
      url='https://github.com/PositroniumSpectroscopy/positronium',
      author='Adam Deller',
      author_email='adam.deller1@gmail.com',
      license='BSD',
      packages=['positronium'],
      install_requires=[
          'scipy>0.14',
      ],
      include_package_data=True,
      zip_safe=False)



#! python
'''
Binding energy of positronium according to calculations by
Richard A. Ferrell (1951) Phys. Rev. 84, 858
http://dx.doi.org/10.1103/PhysRev.84.858

Nb. This does not include the Lamb shift/ radiative corrections.
'''
from __future__ import print_function, division
import numpy as np
from positronium.constants import alpha, au_energy

def epsilon(l, S, J):
    """ scaling of the fine structure shift accoring to l, S and J."""
    epsilon = -0.5 * np.divide(1.0, 2.0*l + 1.0)
    if S == 0:
        # singlet
        pass
    elif S == 1:
        # triplet
        delta = int(l == 0)
        if J == l + 1:
            omega = np.divide(3.0 * l + 4.0,
                              (l + 1.0) * (2.0 * l + 3.0))
        elif J == l:
            omega = -np.divide(1.0, l * (l + 1.0))
        elif J == l - 1:
            omega = -np.divide(3.0 * l - 1.0, l * (2.0 * l - 1.0))
        else:
            raise ValueError("The total angular momentum quantum number 'J' must " + \
                             "be in the range l - 1 < J < l + 1")
        epsilon = epsilon + 7.0 * delta/ 12.0 + np.divide(1 - delta, 8.0*(l + 0.5)) * omega
    else:
        raise ValueError("The total spin quantum number 'S' must be 0 or 1.")
    return epsilon


def fine_structure(n, l, S, J, unit='au'):
    """ fine structure shift (default: atomic units) for quantum numbers n, l, S and J"""
    delta_E = np.power(n, -4.0) * 11.0 * alpha**2 / 64.0 + \
              np.power(n, -3.0) * alpha**2 * epsilon(l, S, J)
    if (unit == 'au') or (unit == 'Hartree'):
            return delta_E
    elif unit in au_energy:
        return au_energy[unit](delta_E)
    else:
        raise KeyError(unit + ' is not recognised as a suitable unit.')

def energy_level(n, l, S, J, unit='au'):
    """ energy level (default: atomic units) of Ps with quantum numbers n, l, S and J"""
    energy = -np.power(n, -2.0) / 4.0 + fine_structure(n, l, S, J, unit='au')
    if (unit == 'au') or (unit == 'Hartree'):
        return energy
    elif unit in au_energy:
        return au_energy[unit](energy)
    else:
        raise KeyError(unit + ' is not recognised as a suitable unit.')